(*----------------------------------------------------------------------------*
 *                                                                            *
 *  Template application - Direct3D9 in VCL                                   *
 *                                                                            *
 *  This template provides simple VCL Delphi application that renders D3D     *
 *  objects in VCL TPanel in windowed mode and can also run render D3D (in    *
 *  special window) then running in fullscreen mode.                          *
 *  Template even allows to move/rotate 'The Teapot' with mouse.              *
 *                                                                            *
 *  Template was written to provide compact starting point then creating      *
 *  Direct3D applications with Delphi (using VCL library)                     *
 *                                                                            *
 *  Copyright (c) 2003 by Alexey Barkovoy                                     *
 *  E-Mail: clootie@reactor.ru                                                *
 *                                                                            *
 *  Modified: 26-Jan-2004                                                     *
 *                                                                            *
 *  Latest version can be downloaded from:                                    *
 *     http://clootie.narod.ru                                                *
 *----------------------------------------------------------------------------*)

unit MainUnit;

interface

uses
  Windows, Messages, Classes, SysUtils,
  Forms, Controls, StdCtrls, ExtCtrls, ComCtrls, Menus, AppEvnts,
  Direct3D9, D3DX9, D3DApp, D3DFont, D3DUtil, DXUtil;

type
  CMyD3DApplication = class;

  TMainForm = class(TForm)
    DisplayPanel: TPanel;
    Splitter1: TSplitter;
    MainMenu1: TMainMenu;
    StatusBar: TStatusBar;
    File1: TMenuItem;
    Exit1: TMenuItem;
    N1: TMenuItem;
    About1: TMenuItem;
    ApplicationEvents1: TApplicationEvents;
    ButtonPanel: TPanel;
    Splitter2: TSplitter;
    FullScreenButton: TButton;
    FullScreen: TMenuItem;
    procedure FormShow(Sender: TObject);
    procedure OnToggleFullscreen(Sender: TObject);
    procedure Exit1Click(Sender: TObject);
    procedure ApplicationEvents1Idle(Sender: TObject; var Done: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure About1Click(Sender: TObject);
    procedure DisplayPanelResize(Sender: TObject);
    procedure FullScreenButtonClick(Sender: TObject);
  private
    { Private declarations }
    fMyApp: CMyD3DApplication;
  public
    { Public declarations }
  end;

  CMyD3DApplication = class(CD3DApplication)
    m_pFont: CD3DFont;        // Font for drawing text
    m_ArcBall: CD3DArcBall;   // Mouse rotation utility
    m_matArcballWorld: TD3DXMatrix;

    fTeapot: ID3DXMesh;
  protected
    function OneTimeSceneInit: HRESULT; override;
    function AdjustWindowForChange: HRESULT; override;
    function InitDeviceObjects: HRESULT; override;
    function RestoreDeviceObjects: HRESULT; override;
    function InvalidateDeviceObjects: HRESULT; override;
    function DeleteDeviceObjects: HRESULT; override;
    function Render: HRESULT; override;
    function FrameMove: HRESULT; override;
    function ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
      adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT; override;
    function FinalCleanup: HRESULT; override;
    function CheckForLostFullscreen: HResult;
  public
    fAppForm: TForm;
    m_hwndRenderWindow:      HWND;
    m_hwndRenderFullScreen:  HWND;
    constructor Create;
    procedure AssignWindowsHandles;
    function MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; override;
  end;

var
  MainForm: TMainForm;

///////////////////////////////////////////////////////////////////////////

const
  sAppName = 'Clootie - VCL Direct3D Template';

implementation


{$R *.dfm}


//-----------------------------------------------------------------------------
// Name: FullScreenWndProc()
// Desc: The WndProc funtion used when the app is in fullscreen mode. This is
//       needed simply to trap the ESC key.
//-----------------------------------------------------------------------------
function FullScreenWndProc(hWnd: HWND; msg: UINT; wParam: WPARAM;
  lParam: LPARAM): LRESULT; stdcall;
begin
  if (msg = WM_CLOSE) then
  begin
    // User wants to exit, so go back to windowed mode and exit for real
    MainForm.OnToggleFullScreen(nil);
    PostMessage(Application.Handle, WM_CLOSE, 0, 0);
  end;

  if (msg = WM_SETCURSOR) then SetCursor(0);

  // User wants to leave fullscreen mode
  if (msg = WM_KEYUP) and (wParam = VK_ESCAPE) then
    MainForm.OnToggleFullScreen(nil);

  if Assigned(MainForm) and Assigned(MainForm.fMyApp) then
    MainForm.fMyApp.m_ArcBall.HandleMessages(hWnd, Msg, wParam, lParam);

  Result:= DefWindowProc(hWnd, msg, wParam, lParam);
end;

procedure TMainForm.FormShow(Sender: TObject);
{$WRITEABLECONST ON}
const
  wndClass: TWndClass = (
    style: CS_HREDRAW or CS_VREDRAW;
    lpfnWndProc: @FullScreenWndProc; cbClsExtra: 0; cbWndExtra: 0; hInstance: 0;
    hIcon: 0; hCursor: 0; hbrBackground: 0; lpszMenuName: nil;
    lpszClassName: 'Fullscreen Window'
  );
{$WRITEABLECONST OFF}
begin
  fMyApp:= CMyD3DApplication.Create;
  fMyApp.fAppForm:= Self;
  fMyApp.m_hwndRenderWindow:= DisplayPanel.Handle; // Save static reference to the render window

  // Register a class for a fullscreen window
  wndClass.hbrBackground:= GetStockObject(WHITE_BRUSH);
  Windows.RegisterClass(wndClass);

  // We create the fullscreen window (not visible) at startup, so it can
  // be the focus window.  The focus window can only be set at CreateDevice
  // time, not in a Reset, so ToggleFullscreen wouldn't work unless we have
  // already set up the fullscreen focus window.
  fMyApp.m_hwndRenderFullScreen:= CreateWindow('Fullscreen Window', nil,
                                    WS_POPUP, 0, 0, Screen.Width, Screen.Height,
                                    Application.Handle, 0, 0, nil);

  fMyApp.AssignWindowsHandles;
  fMyApp.AppCreate(HInstance);
end;

//-----------------------------------------------------------------------------
// Name: OnToggleFullScreen()
// Desc: Called when user toggles the fullscreen mode
//-----------------------------------------------------------------------------
procedure TMainForm.OnToggleFullscreen(Sender: TObject);
begin
  fMyApp.ToggleFullscreen;
end;

procedure TMainForm.Exit1Click(Sender: TObject);
begin
  Close;
end;

procedure TMainForm.ApplicationEvents1Idle(Sender: TObject;
  var Done: Boolean);
var
  strStatsPrev: String;
begin
  Done:= True;

  // Do not render if the app is minimized
  if IsIconic(Application.Handle) then Exit;

  // Update and render a frame
  if Assigned(fMyApp) and fMyApp.m_bActive then
  begin
    strStatsPrev:= fMyApp.m_strFrameStats;
    fMyApp.CheckForLostFullscreen;
    fMyApp.Render3DEnvironment;
    if (strStatsPrev <> fMyApp.m_strFrameStats) then
      StatusBar.Panels[0].Text:= fMyApp.m_strFrameStats;
  end;

  // Keep requesting more idle time
  Done:= False;
end;

procedure TMainForm.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
  DestroyWindow(fMyApp.m_hwndRenderFullScreen);
  fMyApp.m_hwndRenderFullScreen:= 0;

  FreeAndNil(fMyApp);
end;

procedure TMainForm.About1Click(Sender: TObject);
begin
  Application.MessageBox(
    'This demo shows simple template for usage of Direct3D9 with VCL forms.'#10+
    'Demo support resizing of client window at run-time and going to/from fullscreen mode.'#10#10+
    'This and other demos can be found at http://clootie.narod.ru/',
    'Clootie - VCL Direct3D9 template', MB_ICONINFORMATION or MB_OK);
end;

procedure TMainForm.DisplayPanelResize(Sender: TObject);
begin
  if Assigned(fMyApp) and fMyApp.m_bActive then
  begin
    fMyApp.Pause(True);
    fMyApp.MsgProc(0, WM_EXITSIZEMOVE, 0, 0);
  end;
end;

procedure TMainForm.FullScreenButtonClick(Sender: TObject);
begin
  OnToggleFullscreen(nil);
end;


///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Application constructor. Sets attributes for the app.
//-----------------------------------------------------------------------------
constructor CMyD3DApplication.Create;
begin
  inherited;
  m_strWindowTitle  := 'Direct3D Display Panel';
  m_bShowCursorWhenFullscreen := True;
  m_d3dEnumeration.AppUsesDepthBuffer := True;

  m_pFont := CD3DFont.Create('Arial', 12, D3DFONT_BOLD);
  m_ArcBall := CD3DArcBall.Create;
  m_ArcBall.SetTranslationRadius(10);
end;


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
function CMyD3DApplication.OneTimeSceneInit: HRESULT;
begin
  // Set cursor to indicate that user can move the object with the mouse
{$IFDEF WIN64}
  SetClassLongPtr(m_hWnd, GCLP_HCURSOR, LONG_PTR(LoadCursor(0, IDC_SIZEALL)));
{$ELSE}
  SetClassLong(m_hWnd, GCL_HCURSOR, LoadCursor(0, IDC_SIZEALL));
{$ENDIF}

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FrameMove: HRESULT;
begin
  // Update rotation matrix from ArcBall
  m_matArcballWorld:= m_ArcBall.GetRotationMatrix^;
  D3DXMatrixMultiply(m_matArcballWorld, m_matArcballWorld, m_ArcBall.GetTranslationMatrix^);
  m_pd3dDevice.SetTransform(D3DTS_WORLD, m_matArcballWorld);

  Result:= S_OK;
end;



//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
function CMyD3DApplication.Render: HRESULT;
begin
  // Clear the viewport
  m_pd3dDevice.Clear(0, nil, D3DCLEAR_TARGET or D3DCLEAR_ZBUFFER,
                     D3DCOLOR_RGBA(0,0,0,0), 1.0, 0);

  // Begin the scene
  if SUCCEEDED(m_pd3dDevice.BeginScene) then
  begin
    fTeapot.DrawSubset(0);

    // Output statistics
    m_pFont.DrawText(2,  0, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats);
    m_pFont.DrawText(2, 20, D3DCOLOR_ARGB(255,255,255,0), m_strDeviceStats);

    // End the scene
    m_pd3dDevice.EndScene;
  end;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: AdjustWindowForChange()
// Desc: Adjusts the window properties for windowed or fullscreen mode
//-----------------------------------------------------------------------------
function CMyD3DApplication.AdjustWindowForChange: HRESULT;
begin
  if (m_bWindowed) then
  begin
    ShowWindow(m_hwndRenderFullScreen, SW_HIDE);
    m_hWnd:= m_hwndRenderWindow;
  end else
  begin
    if IsIconic(m_hwndRenderFullScreen) then
      ShowWindow(m_hwndRenderFullScreen, SW_RESTORE);

    ShowWindow(m_hwndRenderFullScreen, SW_SHOW);
    m_hWnd:= m_hwndRenderFullScreen;
  end;
  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.InitDeviceObjects:HRESULT;
begin
  // Initialize the font's internal textures
  m_pFont.InitDeviceObjects(m_pd3dDevice);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Initialize scene objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.RestoreDeviceObjects: HRESULT;
var
  fAspect: Single;
  m_matProjection: TD3DXMatrix;
  matView: TD3DXMatrix;
  vEyePt, vLookatPt, vUpVec: TD3DXVector3;
  Light: TD3DLight9;
  Mtrl: TD3DMaterial9;
begin
  m_pFont.RestoreDeviceObjects;
  D3DXCreateTeapot(m_pd3dDevice, fTeapot, nil);
  m_ArcBall.SetWindow(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height);

  // Update projection matrix based on window dimensions
  fAspect := m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height;
  D3DXMatrixPerspectiveFovLH(m_matProjection, D3DX_PI / 4, fAspect, 1.0, 1000.0);
  m_pd3dDevice.SetTransform(D3DTS_PROJECTION, m_matProjection);

  vEyePt    := D3DXVector3(0.0, 1.0, -5.0);
  vLookatPt := D3DXVector3(0.0, 0.0,  0.0);
  vUpVec    := D3DXVector3(0.0, 1.0,  0.0);
  D3DXMatrixLookAtLH(matView, vEyePt, vLookatPt, vUpVec);
  m_pd3dDevice.SetTransform(D3DTS_VIEW, matView);

  D3DUtil_InitLight(Light, D3DLIGHT_DIRECTIONAL, 1, -1, 1);
  Light.Diffuse:= D3DXColor(0, 1, 0.5, 0);
  m_pd3dDevice.SetLight(0, Light);
  m_pd3dDevice.LightEnable(0, True);

  D3DUtil_InitMaterial(Mtrl, 1, 0, 1, 1);
  m_pd3dDevice.SetMaterial(Mtrl);

  m_pd3dDevice.SetRenderState(D3DRS_AMBIENT, $60606060);

  m_pd3dDevice.SetRenderState(D3DRS_LIGHTING, iTrue);
  m_pd3dDevice.SetRenderState(D3DRS_ZENABLE, iTrue);

  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLOROP,   D3DTOP_MODULATE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG1, D3DTA_TEXTURE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_COLORARG2, D3DTA_DIFFUSE);
  m_pd3dDevice.SetTextureStageState(0, D3DTSS_ALPHAOP,   D3DTOP_DISABLE);

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc:
//-----------------------------------------------------------------------------
function CMyD3DApplication.InvalidateDeviceObjects: HRESULT;
begin
  m_pFont.InvalidateDeviceObjects;
  fTeapot:= nil;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
function CMyD3DApplication.DeleteDeviceObjects: HRESULT;
begin
  m_pFont.DeleteDeviceObjects;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
function CMyD3DApplication.ConfirmDevice(const pCaps: TD3DCaps9; dwBehavior: DWORD;
  adapterFormat: TD3DFormat; backBufferFormat: TD3DFormat): HRESULT;
begin
  Result:= E_FAIL;

  // Need to support post-pixel processing (for alpha blending)
  if FAILED(m_pD3D.CheckDeviceFormat(pCaps.AdapterOrdinal, pCaps.DeviceType,
       adapterFormat, D3DUSAGE_RENDERTARGET or D3DUSAGE_QUERY_POSTPIXELSHADER_BLENDING,
       D3DRTYPE_SURFACE, backBufferFormat))
  then Exit;

  // Only D3DFMT_X1R5G5B5, D3DFMT_R5G6B5, or D3DFMT_X8R8G8B8 format
  // are supported for fullscreen GDI mode.
  if not (backBufferFormat in [D3DFMT_X1R5G5B5, D3DFMT_R5G6B5, D3DFMT_X8R8G8B8]) then Exit;

  Result:= S_OK;
end;


//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
function CMyD3DApplication.FinalCleanup: HRESULT;
begin
  SAFE_DELETE(m_pFont);

  Result:= S_OK;
end;

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle user input
//-----------------------------------------------------------------------------
function CMyD3DApplication.MsgProc(hWnd: HWND; uMsg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT;
begin
  // Pass mouse messages to the ArcBall so it can build internal matrices
  m_ArcBall.HandleMessages(hWnd, uMsg, wParam, lParam);
  Result:= inherited MsgProc(hWnd, uMsg, wParam, lParam);
end;

procedure CMyD3DApplication.AssignWindowsHandles;
begin
  // Note that for the VCL samples, the device window and focus window
  // are not the same.
  m_hWnd := m_hwndRenderWindow;
  m_hWndFocus := m_hwndRenderFullScreen;
end;

//-----------------------------------------------------------------------------
// Name: CheckForLostFullscreen()
// Desc: If fullscreen and device was lost (probably due to alt-tab),
//       automatically switch to windowed mode
//-----------------------------------------------------------------------------
function CMyD3DApplication.CheckForLostFullscreen: HResult;
begin
  Result:= S_OK;

  if (m_bWindowed) then Exit;

  if (FAILED(m_pd3dDevice.TestCooperativeLevel)) then
  begin
    ForceWindowed;
    fAppForm.SetFocus;
    Application.BringToFront;
  end;
end;

end.

